#!/usr/bin/perl

#GLOBAL variables
use vars qw/$OUTFILE $CASFILE $baud/;

#
use Getopt::Long qw(:config pass_through permute);
use IO::Handle;
use Pod::Usage;

use constant {
    ZEROBIT => chr(0),
    ONEBIT => chr(255),
    };

use constant {
    START => ZEROBIT,	# start bit is 0
    STOP => ONEBIT,	# stop bit is 1
    MARK => ONEBIT,	# mark signal
    };

use constant {
    ZEROWRITTEN => -2,
    END_OF_FILE => -1,
    RECORD_TOO_SHORT => -3,
    BAUD_SET => -4,
    };

#Subroutines

sub outheader {
    my $samplerate=shift;
    my $titlelen;

    sysread($CASFILE,$buffer,4) or die "Can't read cas header.";
    die "Unrecognized cas header" if ($buffer ne "FUJI");
    sysread($CASFILE,$buffer,2) or die "CAS Read error."; 	# Title length
    $titlelen=unpack("v",$buffer);
    sysread($CASFILE,$buffer,2) or die "CAS Read error."; 	# skip 2 zero bytes 
    if ($titlelen>0) {
	sysread($CASFILE,$buffer,$titlelen) or die "CAS Read error.";
	print "Tape title: $buffer\n";
    }
    print $OUTFILE "RIFF" .					# RIFF chunk header
	    pack("V",0) .					# (file_size-8)
	    "WAVE" .						# RIFF type
	    "fmt " .						# format chunk
	    pack("V",16) .					# chunk data size
	    pack("vv",1,1) .					# no compression, 1 channel
	    pack("VV",$samplerate,$samplerate) .		# sample rate, byte rate
	    pack("vv",1,8) .					# block align, bits per sample, (no extra bytes)
	    "data" .						# data chunk
	    pack("V",0)						# chunk size, (file_size-40)
	    or die "Can't write wav header to output file.";
}

sub outrecord {
    my $buffer;
    my ($block_type,$block_len,$block_aux);
    #my $recordlen;
    my $irglen;
    my $zerolen;
    my $i,$j;
    my @byte;
    my ($samplerate,$forcebaud,$baudmultiplier,$usefirstbaud,$skipshorter)=(shift,shift,shift,shift,shift);
    my $bitlen;
    my $lastbaud;

    $lastbaud=$baud;
    sysread($CASFILE,$buffer,8) or return END_OF_FILE;

    ($block_type,$block_len,$block_aux)=(substr($buffer,0,4),unpack("v",substr($buffer,4,2)),substr($buffer,6,2));
    #print $block_type,",",$block_len,"\n";
     if ($block_len>0) {
	sysread($CASFILE,$buffer,$block_len) or die "CAS read error.";
    }

    if ($block_type eq 'baud') {
	die "Bad data in baud block" if $block_len!=0;
	if ($forcebaud==-1) {
	    $baud=int(unpack("v",$block_aux)*$baudmultiplier);
	} else {
	    $baud=$forcebaud;
	}
	return BAUD_SET;
    } 
    if ($block_type eq 'zero') {				# "long zero"
	die "Bad data in zero block" if $block_len!=2;
	$irglen=int(unpack("v",$block_aux)/1000*$samplerate);	# lenght of irg
	$zerolen=int(unpack("v",$buffer)/1000*$samplerate);	# lenght of zero record
	for($i=1;$i<=$irglen;$i++) {
	    print $OUTFILE ONEBIT or die "Can't write irg to output file.";
	}
	for($i=1;$i<=$zerolen;$i++) {
	    print $OUTFILE ZEROBIT or die "Can't write zero to output file.";
	}
	return ZEROWRITTEN;
    }
    if ($block_type eq 'data') {
	die "Baud rate was not specified before first data block" if $baud==-1;
	# record of data
	$irglen=int(unpack("v",$block_aux)/1000*$samplerate);	# lenght of IRG

	if ((abs($baud-$lastbaud)<30) and $usefirstbaud==1) {
	    $baud=$lastbaud;
	}

	$bitlen=int($samplerate/$baud+0.5);

	for($i=1;$i<=$irglen;$i++) {
	    print $OUTFILE ONEBIT or die "Can't write irg to output file.";
	}

	if ($block_len < $skipshorter) {			# write additional IRG fragment
	    for($i=1;$i<=$block_len*$bitlen;$i++) {		# instead of bytes if skipping short records
		print $OUTFILE ONEBIT or die "Can't write irg to output file.";
	    }
	    return RECORD_TOO_SHORT;
	}

	foreach my $byte_ (split('',$buffer)) {
	    @byte=split('',unpack("b8",$byte_));

	    for($j=1;$j<=$bitlen;$j++) {
		print $OUTFILE START or die "Can't write start bit to output file.";   # start bit
	    }

	    foreach my $bit (@byte) {
		for($j=1;$j<=$bitlen;$j++) {
		    print $OUTFILE chr($bit*0xFF) or die "Can't write data bit to output file."; # data bits
		}
	    }

	    for($j=1;$j<=$bitlen;$j++) {
		print $OUTFILE STOP or die "Can't write stop bit to output file.";   # stop bit
	    }
	}
	return $block_len;
    }
    die "Bad data in CAS file.";
}

sub outtail {		# outputs 10000 samples of mark signal
			# at the end of tape
    my $i;

    for($i=0;$i<=10000;$i++) {
	print $OUTFILE MARK or die "Can't write wav header to output file (tail).";
    }
}

########################
### START PROCESSING ###
########################

my $forcebaud=-1;		# whether we want specific baud rate
my $baudmultiplier=1;		# whether we want multiply original baud by this
my $usefirstbaud=0;		# use first baud value found in cas
my $maxbauddiff=30;		# max baud difference before switching to new "first baud"
my $samplerate=48000;		# sample rate of output file
my $skipshorter=0;		# skip data blocks shorter than this
my $help=0;			# do you want short help ?
my $man=0;			#   or detailed manual ?

my $lastbaud=0;			# baud rate for last seen data block
$baud=-1;			# baud rate of current data block

# save command line options for future reference
open(OPTIONSFILE,">cas2square.options") or print "Can't write to options file.";
print OPTIONSFILE join(" ",@ARGV[0..$#ARGV-1]);
close(OPTIONSFILE);

GetOptions('baud|b=i' => \$forcebaud,
	    'baudmultiplier|m=f' => \$baudmultiplier,
	    'samplerate|r=i' => \$samplerate,
	    'usefirstbaud|f' => \$usefirstbaud,
	    'maxbauddiff|x=i' => \$maxbauddiff,
	    'skipshorter=i'	=> \$skipshorter,
	    'help|h' => \$help,
	    'man' => \$man) or pod2usage(2);

pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;

pod2usage(1) if !@ARGV;

open($CASFILE,"<" . @ARGV[0]) or die "Can't open input file $ARGV[0]";

$ARGV[0]=~s/(.*)(\.cas$)/$1/;

open($OUTFILE,">" . @ARGV[0] . "-rebuilt.wav") or die "Can't open output wav file.";

my $records=0;
my $status=1;
my $longzeros=0;
my $nonstd=0;
my $skippedshort=0;

#STDOUT->autoflush(1);

print "Mastering started.\n";
print "Using ";

if ($forcebaud==-1) {
    print "variable ";
} else {
    print "constant (specified: $forcebaud)";
}

if ($usefirstbaud==1) {
    print "(first significantly different found in cas file) ";
} else {
    print $forcebaud==-1 ? "(specific for each record) ": "";
}

print " baud rate.\n";

outheader($samplerate);

$status=1;

while(true) {
    $status=outrecord($samplerate,$forcebaud,$baudmultiplier,$usefirstbaud,$skipshorter);
    last if $status==END_OF_FILE;
    if ($status==RECORD_TOO_SHORT) {
	$skippedshort++;
	next;
    }
    if ((abs($baud-$lastbaud)>30)) {
	print "Sample rate is $samplerate, baud rate for the block ";
	print eval($records+1) . " and the following blocks will be around " . $baud . ".\n";
	$lastbaud=$baud;
    }
    $records++ if $status>0;
    $longzeros++ if $status==ZEROWRITTEN;
    $nonstd++ if  ($status!=132) and ($status>0);
    print "Blocks written: $records ($nonstd non-standard, $longzeros long zeros, skipped $skippedshort blocks shorter than $skipshorter bytes) ... \r";
    STDOUT->flush;
};

outtail();

print "Blocks written: $records ($nonstd non-standard, $longzeros long zeros, skipped $skippedshort blocks shorter than $skipshorter bytes) ... done.\n";
STDOUT->flush;

# Fix wave header
my $filesize=tell($OUTFILE);
$filesize-=8;

seek $OUTFILE,4,SEEK_SET;
print $OUTFILE pack("i",$filesize);

$filesize-=32;

seek $OUTFILE,40,SEEK_SET;
print $OUTFILE pack("i",$filesize);

# Tadaaaa!
print "Finished.\n";

close($CASFILE);
close($OUTFILE);

1;		# exit nicely

__END__

=head1 NAME

cas2square.pl - converting .cas file to perfect square wave

=head1 SYNOPSIS

cas2square.pl [options] file.cas

  Options (default value):
    --help				this brief help (no)
    --man				full documentation (no)
    --baud|-b <baud>			target baud rate (use cas file)
    --baudmultiplier|-m <multiplier>	multiply baud rate by this (1)
    --samplerate|-r <samplerate>	target sample rate (48000)
    --usefirstbaud|-u			use first baud rate found in cas file (no)
    --maxbauddiff|-x <maxdiff>		maximum difference in baud after which
					new first baud rate is assumed (30)
    --skipshorter <howshort>		skip blocks shorter than these (0)

=head1 OPTIONS

=over 8

=item B<--help>

Print a brief help message and exits.

=item B<--man>

Prints the manual page and exits.

=item B<--baud | -b baud>

Specify baud rate for entire cas file.

=item B<--baudmultiplier | -m multiplier>

Specify multiplier for baud rate saved in .cas file. Use to increase or decrease original baud rate.

=item B<--samplerate | -r samplerate>

Specifies sample rate for output file.

=item B<--usefirstbaud | -u>

The first baud rate found in .cas file will be used for entire file (baud rates for individual records will be ignored).
However, when the difference in baud rate betwen previous and currenttly processed record exceeds 30 
(or value specified in --maxbauddiff option), the new baud rate will be used.

=item B<--maxbauddiff | -x <difference>>

Specifies maximum difference in baud rate which is ignored, when --usefirstbaud is used.
Useful when tape dump contains parts written with different speeds (loader + program).

=item B<--skipshorter <bytes>>

Use it if You want to build wav file skipping obviously unneeded short data records,
which were detected by square2cas.pl and stored in .cas file.

=back

=head1 DESCRIPTION

B<cas2square.pl> will convert .cas file into perfect square wave, which can be used to produce high quality Atari tapes.
Written by B<FUJI> in 2009
Nick 'FUJI' was registered at atari.area (http://atariarea.krap.pl/) in 2009

=cut
