#!/bin/bash

# Atari tape to square wave converter
# Written by FUJI in 2009
# Nick 'FUJI' was registered at atari.area (http://atariarea.krap.pl/) in 2009

# This script uses the following software:
#  ecasound (http://ecasound.seul.org/ecasound/index.php)
#  (LADSPA) SWH plugins (http://plugin.org.uk/)
#  (LADSPA) Blop plugins (http://blop.sourceforge.net/index.html)
#  (LADSPA) CMT plugins (http://www.ladspa.org/cmt/)
#  additionally ecasound uses mpg321 to decode mp3 files

OPT=""
unset MARKFREQ
unset SPACEFREQ
unset BANDWIDTH
AMPLIFY=0

[ "$#" -eq "0" ] && { echo "Parameters are missing..."; exit 1; }

# parse command line
while [ "$#" -ne "1" ]; do
    case $1 in
# 'mark' signal frequency
# rarely needed
    "-mark") { shift; MARKFREQ=$1; shift; }
        ;;
# 'space' signal frequency
# rarely needed
    "-space") { shift; SPACEFREQ=$1; shift; }
        ;;
# resistor - this adds constant signal to the output of "comparator"
# (like in real Atari tape recorder/player).
# This is most usefull command line option.
# The default value will work most of the time for good quality tape dumps.
# You may need to adjust this, if square2cas.pl reports checksum errors in standard data records.
# Values may range from 0.001 to 0.2
    "-resistor") { shift; RESISTOR=$1; shift; }
	;;
# input file may be normalized, sometimes helps
    "-amplify") { AMPLIFY=1; shift; }
	;;
# You may try this if tape dump is of very poor quality
    "-bandwidth") { shift; BANDWIDTH=$1; shift; }
	;;
# ignore anything else until last remains
    *) { shift; }
	;;
    esac
done

[ "$#" -eq "0" ] && { echo "Parameters are BAD..."; exit 1; }

INFILE=${1}

[ -f ${INFILE} ] || { echo "Error opening ${INFILE}."; exit 1; }

MARKFREQ=${MARKFREQ:-5327}
SPACEFREQ=${SPACEFREQ:-3995}
RESISTOR=${RESISTOR:-0.005}
BANDWIDTH=${BANDWIDTH:-1000}

OPTFILE="${INFILE%.*}.options"

# save parameters for future reference
echo "mark freq: ${MARKFREQ}, space freq: ${SPACEFREQ}, resistor: ${RESISTOR}, amplify: ${AMPLIFY}" > ${OPTFILE}

# eliminate harmless glibc error message
export MALLOC_CHECK_=0

CHANNELS=`ecalength -s -f -c ${1} 2>/dev/null`
RATE=`ecalength -s -f -r ${1} 2>/dev/null`
#BITS=`ecalength -s -f -b ${1} 2>/dev/null`

OUTFILE="${INFILE%.*}-squared.wav"

[ "$CHANNELS" -eq "2" ] && { OPT="-f:s16_le,1,${RATE} -chmove:2,1"; }

if [ "${AMPLIFY}" -eq "1" ]; then
    echo -e "Checking, how much the signal can be amplified...\c"
    AMP=`ecasound -i:${1} $OPT -ev -o:null | grep "Max gain" | sed 's/.*clipping:\ \(.*\)\.$/\1/' | awk '{print int($1*100)}'`
echo " $AMP%."
else
    AMP=100
fi

echo -e "Processing...\c"

# I don't want warnings on the screen
exec 2>/dev/null
export MALLOC_CHECK_=0

# All below should be clear, maybe except the following fragment:
#    -el:fbdelay_0.01s,0.00004,0.5,0.4 \
#    -el:fbdelay_0.01s,0.00005,0.5,0.4 \
#    -el:fbdelay_0.01s,0.00006,0.5,0.4 \
# Signal after leaving "comparator" is mixed 3 times with its delayed version.
# Delays are mnimal, 0.04, 0.05, and 0.06 ms.
# It affects only several samples near zero crossing points,
# eliminating ALL unwanted random spikes near signal edges.
#
# Band reject filter is used to eliminate a lot of noise.
#
# Gigantic amplification just before the end of processing
# produces perfect 'binary' square wave.

ecasound -q -x -a:space,mark -i:${INFILE} \
    -a:space $OPT -efr:${MARKFREQ},1000 -el:bandpass_iir,${SPACEFREQ},${BANDWIDTH},3 -ea:${AMP} \
    -el:diode,2 \
    -f:s16_le,2,${RATE} \
    -chmove:1,2 \
    -a:mark $OPT -efr:${SPACEFREQ},1000 -el:bandpass_iir,${MARKFREQ},${BANDWIDTH},3 -ea:${AMP} \
    -el:diode,2 \
    -f:s16_le,2,${RATE} \
    -chmute:2 \
    -z:mixmode,sum \
    -a:space,mark -o:loop,mix \
    -a:diff -i:loop,mix \
    -el:difference_iama_oa \
    -el:fbdelay_0.01s,0.00004,0.5,0.4 \
    -el:fbdelay_0.01s,0.00005,0.5,0.4 \
    -el:fbdelay_0.01s,0.00006,0.5,0.4 \
    -f:u8,1,${RATE} \
    -ezx:1,${RESISTOR} \
    -ea:6553600 -ea:6553600 \
    -o ${OUTFILE}

if [ "$?" -eq "0" ]; then
    echo " finished."
else
    echo " error!"
fi
